"""!calc <equation> will return the google calculator result for <equation>"""
from bs4 import BeautifulSoup
import re
try:
    from urllib import quote
except ImportError:
    from urllib.request import quote
import requests

def calc(eq):
    query = quote(eq)
    url = "https://encrypted.google.com/search?hl=en&q={0}".format(query)
    soup = BeautifulSoup(requests.get(url).text, "html5lib")

    answer = soup.findAll("h2", attrs={"class": "r"})
    if not answer:
        answer = soup.findAll("span", attrs={"class": "_m3b"})
        if not answer:
            return ":crying_cat_face: Sorry, google doesn't have an answer for you :crying_cat_face:"

    # They seem to use u\xa0 (non-breaking space) in place of a comma
    answer = answer[0].text.replace(u"\xa0", ",")
    return answer

def on_message(msg, server):
    text = msg.get("text", "")
    match = re.findall(r"!calc (.*)", text)
    if not match:
        return

    return calc(match[0].encode("utf8"))



"""!nr [-l] chart_name

Show a chart made from new relic data.

To list the available charts, use `!nr -l`
"""
import argparse
from datetime import datetime, timedelta
from dateutil.parser import parse
from io import BytesIO 
import os
from pytz import utc
import re
import requests
import shlex

import matplotlib
# need to call this before importing pyplot
matplotlib.use('Agg')
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pandas as pd
from pandas.io.json import json_normalize
import seaborn as sns

NR_API_KEY = os.environ.get("NR_API_KEY")
NR_QUERY_KEY = os.environ.get("NR_QUERY_KEY")
NR_ACCOUNT_ID = os.environ.get("NR_ACCOUNT_ID")

# TODO set this with an mpl style
sns.set(style="darkgrid")

def dtparse(dt):
    return mdates.date2num(parse(dt))

def ffm_graph(nrql, title):
    url = "https://insights-api.newrelic.com/v1/accounts/{}/query?nrql={}".format(NR_ACCOUNT_ID, nrql)
    r = requests.get(url, headers={
        "Accept": "application/json",
        "X-Query-Key": NR_QUERY_KEY})
    pts = [(mdates.date2num(datetime.fromtimestamp(x['beginTimeSeconds'], utc)),
        x['results'][0]['percentiles']['50'],
        x['results'][0]['percentiles']['90'],
        x['results'][0]['percentiles']['95'],
        x['results'][0]['percentiles']['99'])
        for x in r.json()['timeSeries']]

    df = pd.DataFrame(pts, columns=['time', '50', '90', '95', '99'])
    df = df.set_index('time')

    fig, ax = plt.subplots()
    df.plot(ax=ax, title=title)
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.yaxis.set_label_text("time in ms")
    ax.legend().set_title("percentiles")

    imgbuf = BytesIO()
    img = ax.figure.savefig(imgbuf, format='png')

    imgbuf.seek(0)
    return imgbuf

def ffm_payment():
    nrql = """SELECT percentile(duration, 50, 90, 95, 99)
                FROM Transaction SINCE 30 minutes ago
                WHERE appName = 'Plan Compare 2.0'
                  AND transactionType = 'Other'
                  AND name='OtherTransaction/Background/FFMGetPayment/run'
              TIMESERIES"""
    return ffm_graph(nrql, "FFM Payment Response Times")


def ffm_get():
    nrql = """SELECT percentile(duration, 50, 90, 95, 99)
                FROM Transaction SINCE 30 minutes ago
                WHERE appName = 'Plan Compare 2.0'
                  AND transactionType = 'Other'
                  AND name='OtherTransaction/Background/FFMGet/run'
              TIMESERIES"""
    return ffm_graph(nrql, "FFM Get Response Times")

def ffm_submit():
    nrql = """SELECT percentile(duration, 50, 90, 95, 99)
                FROM Transaction SINCE 30 minutes ago
                WHERE appName = 'Plan Compare 2.0'
                  AND transactionType = 'Other'
                  AND name='OtherTransaction/Background/FFMSubmit/run'
              TIMESERIES"""
    return ffm_graph(nrql, "FFM Submit Response Times")


def response_times():
    nrql = """SELECT percentile(duration, 50, 90, 95, 99)
                FROM Transaction SINCE 30 minutes ago
               WHERE appName = 'Plan Compare 2.0' AND transactionType = 'Web'
              TIMESERIES"""
    url = "https://insights-api.newrelic.com/v1/accounts/{}/query?nrql={}".format(NR_ACCOUNT_ID, nrql)
    r = requests.get(url, headers={
        "Accept": "application/json",
        "X-Query-Key": NR_QUERY_KEY})
    pts = [(mdates.date2num(datetime.fromtimestamp(x['beginTimeSeconds'], utc)),
        x['results'][0]['percentiles']['50'],
        x['results'][0]['percentiles']['90'],
        x['results'][0]['percentiles']['95'],
        x['results'][0]['percentiles']['99'])
        for x in r.json()['timeSeries']]

    df = pd.DataFrame(pts, columns=['time', '50', '90', '95', '99'])
    df = df.set_index('time')
    fig, ax = plt.subplots()
    df.plot(ax=ax, title="Response Times")
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.yaxis.set_label_text("time in ms")
    ax.legend().set_title("percentiles")

    imgbuf = BytesIO()
    img = ax.figure.savefig(imgbuf, format='png')

    imgbuf.seek(0)
    return imgbuf

def count_graph(df, title, dateformat='%H:%M'):
    df = df.set_index('time')
    fig, ax = plt.subplots()
    ax.xaxis_date()
    df.plot(ax=ax, title=title)
    ax.xaxis.set_major_formatter(mdates.DateFormatter(dateformat))
    ax.yaxis.set_label_text("enrollments")

    imgbuf = BytesIO()
    img = ax.figure.savefig(imgbuf, format='png')

    imgbuf.seek(0)
    return imgbuf

def submit_success():
    rpm = requests.get('https://api.newrelic.com/v2/applications/30301056/metrics/data.json',
            headers={'X-Api-Key': NR_API_KEY},
            params={'names[]': 'submit_success', 'values[]': 'call_count'})
    data = rpm.json()
    pts = [(dtparse(x['from']), x['values']['call_count'])
            for x in data["metric_data"]['metrics'][0]['timeslices']]

    df = pd.DataFrame(pts, columns=['time', 'successful_requests'])
    return count_graph(df, "Successful Enrollments")

def submit_error():
    rpm = requests.get('https://api.newrelic.com/v2/applications/30301056/metrics/data.json',
            headers={'X-Api-Key': NR_API_KEY},
            params={'names[]': 'submit_error', 'values[]': 'call_count'})
    data = rpm.json()
    metrics = data["metric_data"]['metrics']
    pts = [(dtparse(x['from']), x['values']['call_count'])
            for x in data["metric_data"]['metrics'][0]['timeslices']]

    df = pd.DataFrame(pts, columns=['time', 'errored_requests'])
    return count_graph(df, "Errored Enrollments")

def ffm_week_over_week():
    nrql = """SELECT average(time)
                FROM PageAction
               WHERE actionName = 'FFMGet'
               SINCE 6 HOURS AGO
              COMPARE WITH 1 WEEK AGO TIMESERIES
    """

    url = "https://insights-api.newrelic.com/v1/accounts/{}/query?nrql={}".format(NR_ACCOUNT_ID, nrql)
    r = requests.get(url, headers={"Accept": "application/json", "X-Query-Key": NR_QUERY_KEY}).json()
    previous = [(d['beginTimeSeconds'], d['results'][0]['average'])
                for d in r['previous']['timeSeries']]
    current = [(d['beginTimeSeconds'], d['results'][0]['average'])
               for d in r['current']['timeSeries']]
    pts = []
    for i, (ts, d) in enumerate(current):
        ts = mdates.date2num(datetime.fromtimestamp(ts, utc))
        pts.append((ts, d, previous[i][1]))

    df = pd.DataFrame(pts, columns=['time', 'current', 'previous'])
    df = df.set_index('time')

    fig, ax = plt.subplots()
    df.plot(ax=ax, title="FFM Get Week over Week")
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.yaxis.set_label_text("time in ms")
    ax.legend().set_title("")

    imgbuf = BytesIO()
    img = ax.figure.savefig(imgbuf, format='png')

    imgbuf.seek(0)
    return imgbuf

def submit_weekly():
    now   = datetime.utcnow()
    from_ = (now - timedelta(7)).isoformat()
    to    = now.isoformat()

    rpm = requests.get('https://api.newrelic.com/v2/applications/30301056/metrics/data.json',
            headers={'X-Api-Key': NR_API_KEY},
            params='names[]=submit_success&names[]=submit_error&values[]=call_count'
                   '&from={}&to={}'.format(from_, to))
    data = rpm.json()

    metrics = data['metric_data']['metrics']

    pts = []
    for i in range(len(metrics[0]['timeslices'])):
        time = dtparse(metrics[0]['timeslices'][i]['from'])
        a = metrics[0]['timeslices'][i]['values']['call_count']
        b = metrics[1]['timeslices'][i]['values']['call_count']
        pts.append((time, a, b))

    df = pd.DataFrame(pts, columns=['time',
        metrics[0]['name'], metrics[1]['name']])

    sum_ = df.sum()
    success, errors = map(int, (sum_['submit_success'], sum_['submit_error']))
    title = 'Enrollments ({} success, {} error)'.format(success, errors)

    return count_graph(df, title, '%b %d')

def submit():
    rpm = requests.get('https://api.newrelic.com/v2/applications/30301056/metrics/data.json',
            headers={'X-Api-Key': NR_API_KEY},
            params='names[]=submit_success&names[]=submit_error&values[]=call_count')
    data = rpm.json()

    metrics = data['metric_data']['metrics']

    pts = []
    for i in range(len(metrics[0]['timeslices'])):
        time = dtparse(metrics[0]['timeslices'][i]['from'])
        a = metrics[0]['timeslices'][i]['values']['call_count']
        b = metrics[1]['timeslices'][i]['values']['call_count']
        pts.append((time, a, b))

    df = pd.DataFrame(pts, columns=['time',
        metrics[0]['name'], metrics[1]['name']])
    return count_graph(df, "Enrollments")

def throughput():
    rpm = requests.get('https://api.newrelic.com/v2/applications/30301056/metrics/data.json',
            headers={'X-Api-Key': NR_API_KEY},
            params={'names[]': 'HttpDispatcher', 'values[]': 'requests_per_minute'})
    data = rpm.json()

    # date2num converts from datetime to matplotlib's particular date format
    # http://matplotlib.org/api/dates_api.html
    pts = [(x['values']['requests_per_minute'], dtparse(x['from']))
            for x in data["metric_data"]['metrics'][0]['timeslices']]
    ds, ts = zip(*pts)

    fig, ax = plt.subplots()
    sns.tsplot(ds, ts, ax=ax, value='Requests Per Minute')
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))

    imgbuf = BytesIO()
    img = fig.savefig(imgbuf, format='png')

    imgbuf.seek(0)
    return imgbuf

charts = {
    "throughput": throughput,
    "response_time": response_times,
    "ffm_get": ffm_get,
    "ffm_submit": ffm_submit,
    "ffm_payment": ffm_payment,
    "success": submit_success,
    "errors": submit_error,
    "submits": submit,
    "submits_weekly": submit_weekly,
    "ffm_get_perf": ffm_week_over_week,
}

def nrq(nrql):
    1/0

def nr(chart):
    return charts.get(chart, charts["throughput"])()

ARGPARSE = argparse.ArgumentParser()
ARGPARSE.add_argument('-l', action='store_true', help="list all available charts")
ARGPARSE.add_argument('-q', help="an NRQL query to run")
ARGPARSE.add_argument('graph', nargs='*', help="the chart to show")

def on_message(msg, server):
    text = msg.get("text", "")
    match = re.findall(r"!nr(.*)?", text)
    if not match:
        return

    if not NR_API_KEY:
        return "New Relic API key not found. Please set the NR_API_KEY " \
               "environment variable to your New Relic API key"

    try:
        ns = ARGPARSE.parse_args(shlex.split(match[0]))
    except SystemExit:
        return __doc__

    if ns.l:
        return "Available charts: {}".format(", ".join(charts.keys()))

    # TODO ns.q
    if ns.q:
        return "not available yet"
        chart = "nrql"
        img = nrq(ns.q)
    else:
        chart = ns.graph[0]
        img = nr(ns.graph[0])

    r = server.slack.api_call("files.upload",
        post_data={"channels": msg["channel"]},
        files={'file': ('{}.png'.format(chart), img, 'image/png')})

    # don't return any messages via RTM
    return ""



# -*- coding: UTF-8 -*-
from collections import namedtuple
import logging
from .mock_handler import MockHandler
import os
import sqlite3
import tempfile

import limbo

# test plugin hooks
#
# TODO: test init_plugins with unicode plugins
# TODO: test init_plugins with invalid plugins
# TODO: test init_plugins with plugin without on_
# TODO: test init_plugins __doc__ handling
# TODO: test plugin that throws exception (on import, init and message)

# copied from slackrtm
User = namedtuple('User', 'server name id real_name tz')

DIR = os.path.dirname(os.path.realpath(__file__))
PARENT = os.path.split(DIR)[0]

os.environ["LIMBO_LOGFILE"] = "/tmp/deleteme"

def test_plugin_success():
    hooks = limbo.init_plugins("test/plugins")
    assert len(hooks) == 7
    assert "message" in hooks
    assert isinstance(hooks, dict)
    assert isinstance(hooks["message"], list)
    assert len(hooks["message"]) == 2

def test_config_plugin_none_success():
    hooks = limbo.init_plugins("test/plugins", None)
    assert len(hooks) == 7
    assert "message" in hooks
    assert isinstance(hooks, dict)
    assert isinstance(hooks["message"], list)
    assert len(hooks["message"]) == 2

def test_config_plugin_empty_string_success():
    hooks = limbo.init_plugins("test/plugins", "")
    assert len(hooks) == 7
    assert "message" in hooks
    assert isinstance(hooks, dict)
    assert isinstance(hooks["message"], list)
    assert len(hooks["message"]) == 2

def test_config_plugin_success():
    hooks = limbo.init_plugins("test/plugins", "echo,loop")
    assert len(hooks) == 5
    assert "message" in hooks
    assert isinstance(hooks, dict)
    assert isinstance(hooks["message"], list)
    assert len(hooks["message"]) == 1

def test_config_plugin_doesnt_exist():
    hooks = limbo.init_plugins("test/plugins", "doesnotexist")
    assert len(hooks) == 0

def test_plugin_invalid_dir():
    try:
        limbo.init_plugins("invalid/package")
    except limbo.InvalidPluginDir:
        return
    1 / 0

def test_plugin_logs():
    mhdr = MockHandler()
    logging.getLogger("limbo.limbo").addHandler(mhdr)
    limbo.init_plugins("test/plugins")
    mhdr.check("debug", "attaching message hook for echo")

# test run_hook

def test_run_hook():
    hooks = limbo.init_plugins("test/plugins")
    assert limbo.run_hook(hooks, "message", {"text": u"!echo bananas"}, None) == [u"!echo bananas"]

def test_missing_hook():
    hooks = limbo.init_plugins("test/plugins")
    assert limbo.run_hook(hooks, "nonexistant", {"text": u"!echo bananas"}, None) == []

# test handle_message

def test_handle_message_subtype():
    msg = u"!echo Iñtërnâtiônàlizætiøn"
    server = limbo.FakeServer()
    event = {"bot_id": "1", "text": msg}
    event["subtype"] = "message_changed"
    assert limbo.handle_message(event, server) == None

def test_handle_message_ignores_self():
    server = limbo.FakeServer()
    event = {"user": "1", "type": "message", "id": "1"}
    assert limbo.handle_message(event, server) == None

def test_handle_message_ignores_slackbot():
    server = limbo.FakeServer()
    event = {"user": "USLACKBOT"}
    assert limbo.handle_message(event, server) == None

def test_handle_message_basic():
    msg = u"!echo Iñtërnâtiônàlizætiøn"
    event = {"user": "2", "text": msg}

    hooks = limbo.init_plugins("test/plugins")
    server = limbo.FakeServer(hooks=hooks)

    assert limbo.handle_message(event, server) == msg

def test_handle_channel_join():
    event = {
      "user": "2",
      "type": "message",
      "subtype": "channel_join",
      "text": "User has joined"
    }

    hooks = limbo.init_plugins("test/plugins")
    server = limbo.FakeServer(hooks=hooks)

    assert limbo.handle_message(event, server) == "saw user 2 join"

# Under unclear circumstances, slack can return a None user.
# https://github.com/llimllib/limbo/issues/40
def test_handle_message_slack_user_nil():
    msg = u"!echo Iñtërnâtiônàlizætiøn"
    event = {"user": "msguser", "text": msg}
    users = {"0": User(None, "nobody", 0, "", 0)}

    hooks = limbo.init_plugins("test/plugins")
    slack = limbo.FakeSlack(users=users)
    server = limbo.FakeServer(slack=slack, hooks=hooks)

    assert limbo.handle_message(event, server) == u"!echo Iñtërnâtiônàlizætiøn"

def test_handle_bot_message():
    msg = u"!echo Iñtërnâtiônàlizætiøn bot"
    event = {"bot_id": "2", "text": msg, "subtype": "bot_message"}

    hooks = limbo.init_plugins("test/plugins")
    server = limbo.FakeServer(hooks=hooks)

    assert limbo.handle_message(event, server) == msg

def test_handle_edit_message():
    oldmsg = u"old message"
    newmsg = u"!echo new message"
    event = {"type": "message",
            "subtype": "message_changed",
            "previous_message": {"text": oldmsg},
            "message": {"text": newmsg, "user": "msguser"}}

    hooks = limbo.init_plugins("test/plugins")
    server = limbo.FakeServer(hooks=hooks)

    assert limbo.handle_message(event, server) == "Changed: {}".format(newmsg)

def test_handle_delete_message():
    oldmsg = u"!echo old message"
    event = {"type": "message",
            "subtype": "message_deleted",
            "previous_message": {"text": oldmsg, "user": "msguser"}}

    hooks = limbo.init_plugins("test/plugins")
    server = limbo.FakeServer(hooks=hooks)

    assert limbo.handle_message(event, server) == "Deleted: {}".format(oldmsg)

def test_init_db():
    tf = tempfile.NamedTemporaryFile()
    db = limbo.init_db(tf.name)
    assert type(db) == type(sqlite3.connect(":memory:"))

class FakeSlackClient(object):
    def __init__(self, connect=True):
        self.connect = connect

    def rtm_connect(self):
        return self.connect

def test_loop_hook():
    hooks = limbo.init_plugins("test/plugins")
    server = limbo.FakeServer(hooks=hooks)
    slack = limbo.FakeSlack()
    limbo.loop(server, test_loop=1)

    assert server._loop_plugin_ran == True
